/*
   TODO:
	- enable brownout @2.7v
*/
#include <avr/io.h>
#include <avr/interrupt.h>
/* #include <avr/eeprom.h> */
#include <avr/power.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <stdlib.h>
#define F_CPU 9600000
/* #include <util/delay.h> */

#define M_WHEEL_ACTIVE 1
#define M_MAIN_MENU 2
#define M_SUB_MENU 4
#define M_AUTO_MOVE 8
#define M_AUTO_MOVE_REV 16
#define M_DOUBLE_RANGE 32

#define PWM_PERIOD_250 (F_CPU/8/250)
#define PWM_PERIOD_333 (F_CPU/8/333)
#define PWM_PERIOD_80 (F_CPU/8/80)
#define PWM_PERIOD_50 (F_CPU/8/50)
#define uS2TICKS 1.2 /* F_CPU/8/1000000 */
#define MIDPULSE (uint16_t)(1520*uS2TICKS)
#define RANGE (1024+1024)
#define MINPULSE (MIDPULSE-RANGE/2)
#define MAXPULSE (MIDPULSE+RANGE/2)

#define RED_LED (1<<DDB3)
#define GREEN_LED (1<<DDB4)
#define KEY (1<<DDB1)
#define OUT (1<<DDB0)

uint8_t cnt_step;
uint16_t cnt;
volatile uint16_t pulse;
uint16_t pause;

volatile uint8_t tick;
/* uint16_t tick1; */
volatile uint16_t pwm_period;

inline uint8_t calc_step(uint16_t p) {
 p -= 1;
 while(p >= 0x100) p>>=1;
 return p;
}

ISR(WDT_vect) {
 tick++;
}

ISR(TIM0_COMPA_vect) {
  uint16_t cntx = cnt-1;
  if(cntx > cnt_step) {
   OCR0A = cnt_step;
   cntx = cntx - cnt_step;
  } else {
   OCR0A = cntx;
   if(PINB & OUT) {
	/* prepare for pause */
    cntx = pause;
    TCCR0A &=  ~(1<<COM0A0);
   } else {
	/* prepare for pulse */
	cntx = pulse;
	pause = pwm_period - cntx;
	TCCR0A |= (1<<COM0A1) | (1<<COM0A0);
   }
   cnt_step = calc_step(cntx);
  }
  cnt = cntx;
/*
  cntx = tick1;
  cntx += OCR0A;
  if(cntx > 10000) {tick++; cntx -= 10000;}
  tick1 = cntx; */
}


static __inline__ uint16_t
readadc(void)
{
	
	uint16_t __ret;

	__asm__ __volatile__ (
		"in %A0,4" "\n\t"
		"in %B0,5" "\n\t"
		"com %B0" "\n\t"
		"com %A0" "\n\t"
		"andi %A0,0xc0" "\n\t"
		: "=d" (__ret)
		:
		: "cc"
	);
	return __ret;
}

static uint16_t adc2p(uint16_t v, uint8_t mode)
{
 if(mode & M_DOUBLE_RANGE) v+=v;
 else v+=RANGE/4;
 v+=MINPULSE;
 return v;
}

static inline uint8_t wheel_controls_pulse(uint8_t mode, uint16_t p/*adcval*/, 
				int16_t end1, int16_t end2, int16_t *curpos) {
 if(mode & M_AUTO_MOVE) {
  if(mode & M_AUTO_MOVE_REV) *curpos -= p>>4;
  else *curpos += p>>4;
 } else *curpos = p;

 if((mode & M_SUB_MENU) == 0) {
  if((int16_t)*curpos > end2) {
   *curpos = end2;
    mode |= M_AUTO_MOVE_REV;
  }
  if(*curpos < end1) {
   *curpos = end1;
   mode &= ~M_AUTO_MOVE_REV;
  }
 }

 p = adc2p(*curpos>>4, mode);
 cli();
 pulse = p;
 sei();
 return mode;
}

/*
 * RED_LED:   *=blink
 *            <btn> * <wheel> center
 *            <btn> ** <wheel> freq selection (<wheel> to exit to manual mode)
 *                              <btn> *    50Hz
 *                              <btn> **   80Hz
 *                              <btn> ***  250Hz
 *                              <btn> **** 333Hz
 *            <btn> *** <wheel> endpoint 1 selection
 *                      <btn> endpoint 2 selection <btn> exit to manual mode
 *            <btn> **** <wheel> auto move
 *            <btn> ***** <wheel> toggle double range mode
 *            <btn> manual mode, exit menu
 *            6*  error: direct jump to reset vector <btn> exit to manual mode
 *            ?*  error: watchdog reset <btn> exit to manual mode
 *            ?*  error: brownout reset <btn> exit to manual mode
 *            ?*  error: reset, unknown cause <btn> exit to manual mode
 */

inline uint16_t process_green_led(uint8_t mode, uint8_t green_led_t, uint16_t pulse) {
 if(!(mode & M_WHEEL_ACTIVE)) {
  if(mode & M_SUB_MENU) PORTB |= GREEN_LED;
  else PORTB &= ~GREEN_LED;
  return 0;
 }
 if(pulse < MIDPULSE + 15 && pulse > MIDPULSE - 15) {
  PORTB |= GREEN_LED;
  return 0;
 }
 if(green_led_t != 0) {
  green_led_t--;
  return green_led_t;
 }
 if(PINB & GREEN_LED) {
  green_led_t = (pwm_period-pulse)>>8;
 } else {
  green_led_t = pulse>>8;
 }

 PINB |= GREEN_LED;
 return green_led_t;
}

static inline uint8_t process_red_led(uint8_t mode, uint8_t selected_menu, 
                      uint8_t led_delay, uint8_t *led_st) {
 if((mode & M_MAIN_MENU) == 0) {
  PORTB &= ~RED_LED;
  return 0;
 }
 selected_menu &= 0xf;
 if((((selected_menu<<4) ^ *led_st) & 0xf0) != 0) {
  *led_st = selected_menu << 4;
 } else {
  if(led_delay != 0) {led_delay--; return led_delay;}
  if(!(PINB & RED_LED)) {
   PORTB |= RED_LED;
   return 20;
  }
 }
 PORTB &= ~RED_LED;
 if((*led_st & 0xf) == 0) {
  *led_st |= *led_st >> 4;
  return 90;
 }
 (*led_st)--;
 return 30;
}

static inline uint8_t wheel_controls_msel(uint8_t mode, uint8_t adcval,
				                          uint8_t *selected_menu,
										  uint8_t *menu_start_adc)
{ 
 if(abs(adcval - *menu_start_adc) < 10) return mode;
 if(mode & M_SUB_MENU) {
  if((*selected_menu & 0xf0) != 0) return mode; /* continue endpoint selection */
  return (mode & M_DOUBLE_RANGE)|M_WHEEL_ACTIVE; /* exit freq selection menu */
 }
 if(*selected_menu == 0) {
  cli();
  pulse = MIDPULSE;
  sei();
  return mode & M_DOUBLE_RANGE; /* hold mid pulse, exit menu */
 }
 if(*selected_menu == 1) {
  uint8_t pp = pwm_period>>8;
  if(pp == (PWM_PERIOD_50>>8)) *selected_menu = 0;
  else if(pp == (PWM_PERIOD_80>>8)) *selected_menu = 1;
  else if(pp == (PWM_PERIOD_250>>8)) *selected_menu = 2;
  else *selected_menu = 3;
  *menu_start_adc = adcval;
  return mode | M_SUB_MENU; /* enter freq menu */
 }
 if(*selected_menu == 2) {
  *selected_menu = 0x10; /* 1 = endpoint1, 2 = endpoint2 */
  return mode | M_SUB_MENU | M_WHEEL_ACTIVE; /* endpoint selection */
 }
 if(*selected_menu == 3)
  return (mode & M_DOUBLE_RANGE)|M_AUTO_MOVE|M_WHEEL_ACTIVE; /* start auto move through range, exit menu */

 if(*selected_menu == 4)
  return ((mode ^ M_DOUBLE_RANGE) & M_DOUBLE_RANGE)|M_WHEEL_ACTIVE;

 return mode;
}

static inline uint8_t process_key(uint8_t mode, uint16_t adcval, uint8_t adcval_div4, uint8_t *selected_menu, 
 uint16_t *end1, uint16_t *end2, uint8_t *menu_start_adc)
{
 uint16_t freq;
 if(mode & M_MAIN_MENU) {
   if((mode & M_SUB_MENU) != 0) {
    uint8_t sm = *selected_menu & 0xf0;
    if(sm == 0) {
     if(*selected_menu == 0) freq = PWM_PERIOD_80;
     else if(*selected_menu == 1) freq = PWM_PERIOD_250;
     else if(*selected_menu == 2) freq = PWM_PERIOD_333;
     else {freq = PWM_PERIOD_50; *selected_menu = 0xff;}
     cli();
     pwm_period = freq;
     sei();
    } else {
	 if(*selected_menu == 0x10) {
	  *end1 = adcval;
	 } else {
	  if(adcval > *end1) *end2 = adcval;
	  else {
	   *end2 = *end1;
	   *end1 = adcval;
	  }
	  return (mode&M_DOUBLE_RANGE)|M_WHEEL_ACTIVE;
     }
    }
   } else if(*selected_menu >= 4) return (mode&M_DOUBLE_RANGE)|M_WHEEL_ACTIVE;
   (*selected_menu)++;
   return mode;
 }
 *menu_start_adc = adcval_div4;
 *selected_menu = 0;
 return (mode&M_DOUBLE_RANGE)|M_MAIN_MENU;
}

void main () __attribute__ ((noreturn));

void main() {
 uint8_t mode = M_WHEEL_ACTIVE;
 uint8_t  greenled = 0, redled;
 uint8_t selected_menu;
 uint16_t end1=0, end2=0x3ff0;
 uint8_t menu_start_adc;
 int16_t curpos;
 uint8_t keyint;
 uint8_t led_st = 0;

 uint8_t sr;
 sr = MCUSR;
 if(sr != (1<<PORF)) {
 #if 0
  selected_menu = 9;
  if(sr == 0) selected_menu = 5;
  if(sr == (1<<WDRF)) selected_menu = 6;
  if(sr == (1<<BORF)) selected_menu = 7;
  if(sr == (1<<EXTRF)) selected_menu = 8;
 #else
  if(sr & (1<<PORF)) selected_menu = 15;
  else selected_menu = (sr>>1)+5;
 #endif
  mode |= M_MAIN_MENU;
 }
 MCUSR = 0;
 DDRB = OUT|RED_LED|GREEN_LED;
 clock_prescale_set(0);
 /*OSCCAL = 0x64;*/
 ADMUX = (1<<MUX0) | (1<<ADLAR); /* PB2=ADC1 */
 ADCSRA = (1<<ADSC) | (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1);
 TCCR0A = 1<<WGM01;
 TIMSK0 = 1<<OCIE0A;
 ACSR = ACD;
 DIDR0 = 1<<ADC1D;
 PORTB = KEY; /* enable pullup for key */
 MCUCR = (MCUCR & ~(1<<ISC00)) | (1<<ISC01); /* INT0, falling edge */
 pwm_period = PWM_PERIOD_50;
 wdt_reset();
 WDTCR = (1<<WDCE)|(1<<WDE);
 WDTCR = (1<<WDTIE)|(1<<WDE);
 sei();

 TCCR0B = 1<<CS01; /* start generator */
 while(1) {
  uint8_t adcval_div4;
  uint16_t adcval_x16;
  uint16_t adcval_x64 = readadc();
  adcval_div4 = adcval_x64>>8;
  adcval_x16 = adcval_x64 >> 2;
   
  if(GIFR & (1<<INTF0)) {
	keyint = 1;
	GIFR = 1<<INTF0; /* clear int0 flag */
	/* wait 10-20 ms for debouncing */
  } else if(!(PINB & KEY) && (keyint&1)) {
     mode = process_key(mode, adcval_x16, adcval_div4, &selected_menu, 
					    &end1, &end2, &menu_start_adc);
	 keyint = 0;
  }
  if(mode & M_WHEEL_ACTIVE) mode = wheel_controls_pulse(mode, adcval_x16, end1, end2, &curpos);
  if(mode & M_MAIN_MENU) mode = wheel_controls_msel(mode, adcval_div4,
				  &selected_menu, &menu_start_adc);
  redled = process_red_led(mode, selected_menu, redled, &led_st);
  greenled = process_green_led(mode, greenled, pulse);

  while(1) {
   WDTCR |= 1<<WDTIE;
   if(tick) {
	cli(); tick--; sei();
	break;
   }
   sleep_mode();
  }
 }
}
